import unittest

from flumine import config


class ConfigTest(unittest.TestCase):
    def test_init(self):
        self.assertFalse(config.simulated)
        self.assertIsInstance(config.hostname, str)
        self.assertIsInstance(config.process_id, int)
        self.assertIsNone(config.current_time)
        self.assertFalse(config.raise_errors)
        self.assertEqual(config.max_execution_workers, 32)
        self.assertFalse(config.async_place_orders)



import logging
import datetime
from collections import defaultdict

from ..baseflumine import BaseFlumine
from ..events import events
from .. import config, utils
from ..clients import ExchangeType
from ..exceptions import RunError
from ..order.trade import TradeStatus
from ..order import process

logger = logging.getLogger(__name__)


class FlumineBacktest(BaseFlumine):
    """
    Single threaded implementation of flumine
    for backtesting strategies with betfair
    historic (or self recorded) streaming data.
    """

    BACKTEST = True

    def __init__(self, client):
        super(FlumineBacktest, self).__init__(client)
        self.handler_queue = []

    def run(self) -> None:
        if self.client.EXCHANGE != ExchangeType.SIMULATED:
            raise RunError(
                "Incorrect client provided, only a Simulated client can be used when backtesting"
            )
        with self:
            self._monkey_patch_datetime()
            """
            list of either single stream or complete events depending 
            on event_processing flag:
               single: {None: [<Stream 1>], [<Stream 2>, ..]}
               event: {123: [<Stream 1>, <Stream 2>, ..], 456: [..]}
            Event data to be muxed/processed chronologically as per 
            live rather than single which is per market in isolation.
            """
            event_streams = defaultdict(list)  # eventId: [<Stream>, ..]
            for stream in self.streams:
                event_id = stream.event_id if stream.event_processing else None
                event_streams[event_id].append(stream)

            for event_id, streams in event_streams.items():
                if event_id and len(streams) > 1:
                    logger.info(
                        "Starting historical event '{0}'".format(event_id),
                        extra={
                            "event_id": event_id,
                            "markets": [s.market_filter for s in streams],
                        },
                    )
                    # create cycles
                    cycles = []  # [[epoch, [MarketBook], gen], ..]
                    for stream in streams:
                        stream_gen = stream.create_generator()()
                        market_book = next(stream_gen)
                        publish_time_epoch = market_book[0].publish_time_epoch
                        cycles.append([publish_time_epoch, market_book, stream_gen])
                    # process cycles
                    while cycles:
                        # order by epoch
                        cycles.sort(key=lambda x: x[0])
                        # get current
                        _, market_book, stream_gen = cycles.pop(0)
                        # process current
                        self._process_market_books(events.MarketBookEvent(market_book))
                        # gen next
                        try:
                            market_book = next(stream_gen)
                        except StopIteration:
                            continue
                        publish_time_epoch = market_book[0].publish_time_epoch
                        # add back
                        cycles.append([publish_time_epoch, market_book, stream_gen])
                    self.handler_queue.clear()
                    logger.info("Completed historical event '{0}'".format(event_id))
                else:
                    for stream in streams:
                        logger.info(
                            "Starting historical market '{0}'".format(
                                stream.market_filter
                            ),
                            extra={
                                "market": stream.market_filter,
                            },
                        )
                        stream_gen = stream.create_generator()
                        for event in stream_gen():
                            self._process_market_books(events.MarketBookEvent(event))
                        self.handler_queue.clear()
                        logger.info(
                            "Completed historical market '{0}'".format(
                                stream.market_filter
                            )
                        )

            self._process_end_flumine()

            logger.info("Backtesting complete")

            self._unpatch_datetime()

    def _process_market_books(self, event: events.MarketBookEvent) -> None:
        # todo DRY!
        for market_book in event.event:
            market_id = market_book.market_id
            config.current_time = market_book.publish_time

            # check if there are orders to process (limited to current market only)
            if self.handler_queue:
                self._check_pending_packages(market_id)

            if market_book.status == "CLOSED":
                self._process_close_market(event=events.CloseMarketEvent(market_book))
                continue

            # get market
            market = self.markets.markets.get(market_id)
            if market is None:
                market = self._add_market(market_id, market_book)
                self.log_control(events.MarketEvent(market))
            elif market.closed:
                self.markets.add_market(market_id, market)

            # process market
            market(market_book)

            # process middleware
            for middleware in self._market_middleware:
                utils.call_middleware_error_handling(middleware, market)

            # process current orders
            self._process_backtest_orders(market)

            for strategy in self.strategies:
                if utils.call_strategy_error_handling(
                    strategy.check_market, market, market_book
                ):
                    utils.call_strategy_error_handling(
                        strategy.process_market_book, market, market_book
                    )

    def process_order_package(self, order_package) -> None:
        # place in pending list (wait for latency+delay)
        self.handler_queue.append(order_package)

    def _process_backtest_orders(self, market) -> None:
        """Remove order from blotter live
        orders if complete and process
        orders through strategies
        """
        blotter = market.blotter
        for order in blotter.live_orders:
            process.process_current_order(order)
            if order.trade.status == TradeStatus.COMPLETE:
                blotter.complete_order(order)
        for strategy in self.strategies:
            strategy_orders = blotter.strategy_orders(strategy)
            utils.call_process_orders_error_handling(strategy, market, strategy_orders)

    def _check_pending_packages(self, market_id: str) -> None:
        processed = []
        for order_package in self.handler_queue:
            if (
                order_package.market_id == market_id
                and order_package.elapsed_seconds > order_package.simulated_delay
            ):
                order_package.client.execution.handler(order_package)
                processed.append(order_package)
        for p in processed:
            self.handler_queue.remove(p)

    def _monkey_patch_datetime(self) -> None:
        config.current_time = datetime.datetime.utcnow()

        class NewDateTime(datetime.datetime):
            @classmethod
            def utcnow(cls):
                return config.current_time

        self._old_datetime = datetime.datetime
        datetime.datetime = NewDateTime

    def _unpatch_datetime(self) -> None:
        datetime.datetime = self._old_datetime

    def __repr__(self) -> str:
        return "<FlumineBacktest>"

    def __str__(self) -> str:
        return "<FlumineBacktest>"



import logging
import unittest
from unittest import mock

from flumine import utils, FlumineException


class UtilsTest(unittest.TestCase):
    def setUp(self) -> None:
        logging.disable(logging.CRITICAL)

    def test_create_short_uuid(self):
        self.assertTrue(utils.create_short_uuid())

    def test_file_line_count(self):
        self.assertGreater(utils.file_line_count(__file__), 10)

    def test_get_file_event_id(self):
        self.assertEqual(
            utils.get_file_md("tests/resources/PRO-1.170258213", "eventId"), "29761984"
        )

    def test_get_file_event_id_tuple(self):
        self.assertEqual(
            utils.get_file_md(("tests/resources/PRO-1.170258213", "test"), "eventId"),
            "29761984",
        )

    def test_chunks(self):
        self.assertEqual([i for i in utils.chunks([1, 2, 3], 1)], [[1], [2], [3]])

    def test_create_cheap_hash(self):
        self.assertEqual(
            utils.create_cheap_hash("test"),
            utils.create_cheap_hash("test"),
        )
        self.assertEqual(len(utils.create_cheap_hash("test", 16)), 16)

    def test_as_dec(self):
        utils.as_dec(2.00)

    # def test_arrange(self):
    #     utils.arange()

    def test_make_prices(self):
        prices = utils.make_prices(utils.MIN_PRICE, utils.CUTOFFS)
        self.assertEqual(len(prices), 350)

    def test_get_nearest_price(self):
        self.assertEqual(utils.get_nearest_price(1.011), 1.01)
        self.assertEqual(utils.get_nearest_price(0), 1.01)
        self.assertEqual(utils.get_nearest_price(1001), 1000)
        self.assertEqual(utils.get_nearest_price(2.01), 2.02)
        self.assertEqual(utils.get_nearest_price(2.0099), 2.00)

    def test_get_price(self):
        self.assertEqual(
            utils.get_price(
                [{"price": 12, "size": 120}, {"price": 34, "size": 120}], 0
            ),
            12,
        )
        self.assertEqual(
            utils.get_price(
                [{"price": 12, "size": 120}, {"price": 34, "size": 120}], 1
            ),
            34,
        )
        self.assertIsNone(
            utils.get_price([{"price": 12, "size": 120}, {"price": 34, "size": 120}], 3)
        )
        self.assertIsNone(utils.get_price([], 3))

    def test_get_size(self):
        self.assertEqual(
            utils.get_size([{"price": 12, "size": 12}, {"price": 34, "size": 34}], 0),
            12,
        )
        self.assertEqual(
            utils.get_size([{"price": 12, "size": 12}, {"price": 34, "size": 34}], 1),
            34,
        )
        self.assertIsNone(
            utils.get_size([{"price": 12, "size": 12}, {"price": 34, "size": 34}], 3)
        )
        self.assertIsNone(utils.get_size([], 3))

    def test_get_sp(self):
        mock_runner = mock.Mock()
        mock_runner.sp = []
        self.assertIsNone(utils.get_sp(mock_runner))
        mock_runner.sp = None
        self.assertIsNone(utils.get_sp(mock_runner))
        mock_runner = mock.Mock()
        mock_runner.sp.actual_sp = "NaN"
        self.assertIsNone(utils.get_sp(mock_runner))
        mock_runner.sp.actual_sp = 12.2345
        self.assertEqual(utils.get_sp(mock_runner), 12.2345)

    def test_price_ticks_away(self):
        self.assertEqual(utils.price_ticks_away(1.01, 1), 1.02)
        self.assertEqual(utils.price_ticks_away(1.01, 5), 1.06)
        self.assertEqual(utils.price_ticks_away(500, 1), 510)
        self.assertEqual(utils.price_ticks_away(500, -1), 490)
        self.assertEqual(utils.price_ticks_away(1.01, -1), 1.01)
        self.assertEqual(utils.price_ticks_away(1.10, -10), 1.01)
        self.assertEqual(utils.price_ticks_away(1000, 5), 1000)
        with self.assertRaises(ValueError):
            utils.price_ticks_away(999, -1)

    def test_calculate_matched_exposure(self):
        self.assertEqual(utils.calculate_matched_exposure([], []), (0.0, 0.0))
        self.assertEqual(
            utils.calculate_matched_exposure([(0, 0)], [(0, 0), (0, 0)]), (0.0, 0.0)
        )
        self.assertEqual(utils.calculate_matched_exposure([(5.6, 2)], []), (9.2, -2.0))
        self.assertEqual(utils.calculate_matched_exposure([], [(5.6, 2)]), (-9.2, 2.0))
        self.assertEqual(
            utils.calculate_matched_exposure([], [(5.6, 2), (5.8, 2)]), (-18.8, 4.0)
        )
        self.assertEqual(
            utils.calculate_matched_exposure([(5.6, 2)], [(5.6, 2)]), (0.0, 0.0)
        )
        self.assertEqual(
            utils.calculate_matched_exposure([(5.6, 2), (100, 20)], [(5.6, 2)]),
            (1980.0, -20.0),
        )
        self.assertEqual(
            utils.calculate_matched_exposure([(5.6, 2), (100, 20)], [(10, 1000)]),
            (-7010.80, 978.0),
        )
        self.assertEqual(
            utils.calculate_matched_exposure([(10, 2)], [(5, 2)]), (10.0, 0.0)
        )
        self.assertEqual(
            utils.calculate_matched_exposure([(10, 2)], [(5, 4)]), (2.0, 2.0)
        )
        self.assertEqual(
            utils.calculate_matched_exposure([(10, 2)], [(5, 8)]), (-14.0, 6.0)
        )

        self.assertEqual(
            utils.calculate_matched_exposure([(5.6, 200)], [(5.6, 100)]),
            (460.0, -100.0),
        )

    def test_calculate_unmatched_exposure(self):
        self.assertEqual(utils.calculate_unmatched_exposure([], []), (0.0, 0.0))
        self.assertEqual(
            utils.calculate_unmatched_exposure([(0, 0)], [(0, 0), (0, 0)]), (0.0, 0.0)
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure([(5.6, 2.0)], []), (0.0, -2.0)
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure([], [(5.6, 2)]), (-9.2, 0.0)
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure([], [(5.6, 2), (5.8, 2)]), (-18.8, 0.0)
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure([(5.6, 2)], [(5.6, 2)]), (-9.2, -2.0)
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure([(5.6, 2), (100, 20)], [(5.6, 2)]),
            (-9.2, -22.0),
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure(
                [(5.6, 2.0), (100.0, 20.0)], [(10, 1000)]
            ),
            (-9000.0, -22.0),
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure([(10.0, 2.0)], [(5, 2)]), (-8.0, -2.0)
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure([(10, 2)], [(5, 4)]), (-16.0, -2.0)
        )
        self.assertEqual(
            utils.calculate_unmatched_exposure([(10, 2)], [(5, 8)]), (-32.0, -2.0)
        )

        self.assertEqual(
            utils.calculate_unmatched_exposure([(5.6, 200.0)], [(5.6, 100)]),
            (-460.0, -200.0),
        )

    def test_wap(self):
        self.assertEqual(
            utils.wap([(123456789, 1.5, 100), (123456789, 1.6, 100)]), (200, 1.55)
        )
        self.assertEqual(utils.wap([]), (0, 0))
        self.assertEqual(utils.wap([(123456789, 1.5, 0)]), (0, 0))

    def test_call_strategy_error_handling(self):
        mock_strategy_check = mock.Mock()
        mock_market = mock.Mock()
        mock_market_book = mock.Mock()
        utils.call_strategy_error_handling(
            mock_strategy_check, mock_market, mock_market_book
        )
        mock_strategy_check.assert_called_with(mock_market, mock_market_book)

    def test_call_strategy_error_handling_flumine_error(self):
        mock_strategy_check = mock.MagicMock(side_effect=FlumineException)
        mock_strategy_check.__name__ = "mock_strategy_check"
        mock_market = mock.Mock()
        mock_market_book = mock.Mock()
        self.assertFalse(
            utils.call_strategy_error_handling(
                mock_strategy_check, mock_market, mock_market_book
            )
        )
        mock_strategy_check.assert_called_with(mock_market, mock_market_book)

    def test_call_strategy_error_handling_error(self):
        mock_strategy_check = mock.MagicMock(side_effect=ValueError)
        mock_strategy_check.__name__ = "mock_strategy_check"
        mock_market = mock.Mock()
        mock_market_book = mock.Mock()
        self.assertFalse(
            utils.call_strategy_error_handling(
                mock_strategy_check, mock_market, mock_market_book
            )
        )
        mock_strategy_check.assert_called_with(mock_market, mock_market_book)

    @mock.patch("flumine.utils.config")
    def test_call_strategy_error_handling_raise(self, mock_config):
        mock_config.raise_errors = True
        mock_strategy_check = mock.MagicMock(side_effect=ValueError)
        mock_strategy_check.__name__ = "mock_strategy_check"
        mock_market = mock.Mock()
        mock_market_book = mock.Mock()
        with self.assertRaises(ValueError):
            utils.call_strategy_error_handling(
                mock_strategy_check, mock_market, mock_market_book
            )

    def test_call_middleware_error_handling(self):
        mock_middlware = mock.Mock()
        mock_market = mock.Mock()
        utils.call_middleware_error_handling(mock_middlware, mock_market)
        mock_middlware.assert_called_with(mock_market)

    def test_call_middleware_error_handling_flumine_error(self):
        mock_middlware = mock.MagicMock(side_effect=FlumineException)
        mock_market = mock.Mock()
        utils.call_middleware_error_handling(mock_middlware, mock_market)
        mock_middlware.assert_called_with(mock_market)

    def test_call_middleware_error_handling_error(self):
        mock_middlware = mock.MagicMock(side_effect=ValueError)
        mock_market = mock.Mock()
        utils.call_middleware_error_handling(mock_middlware, mock_market)
        mock_middlware.assert_called_with(mock_market)

    @mock.patch("flumine.utils.config")
    def test_call_middleware_error_handling_raise(self, mock_config):
        mock_config.raise_errors = True
        mock_middlware = mock.MagicMock(side_effect=ValueError)
        mock_market = mock.Mock()
        with self.assertRaises(ValueError):
            utils.call_middleware_error_handling(mock_middlware, mock_market)

    def test_call_process_orders_error_handling(self):
        mock_strategy = mock.Mock()
        mock_market = mock.Mock()
        utils.call_process_orders_error_handling(mock_strategy, mock_market, [])
        mock_strategy.process_orders.assert_called_with(mock_market, [])

    def test_call_process_orders_error_handling_flumine_error(self):
        mock_strategy = mock.MagicMock()
        mock_strategy.process_orders.side_effect = FlumineException
        mock_market = mock.Mock()
        utils.call_process_orders_error_handling(mock_strategy, mock_market, [])
        mock_strategy.process_orders.assert_called_with(mock_market, [])

    def test_call_process_orders_error_handling_error(self):
        mock_strategy = mock.MagicMock()
        mock_strategy.process_orders.side_effect = ValueError
        mock_market = mock.Mock()
        utils.call_process_orders_error_handling(mock_strategy, mock_market, [])
        mock_strategy.process_orders.assert_called_with(mock_market, [])

    @mock.patch("flumine.utils.config")
    def test_call_process_orders_error_handling_raise(self, mock_config):
        mock_config.raise_errors = True
        mock_strategy = mock.MagicMock()
        mock_strategy.process_orders.side_effect = ValueError
        mock_market = mock.Mock()
        with self.assertRaises(ValueError):
            utils.call_process_orders_error_handling(mock_strategy, mock_market, [])

    def test_get_runner_book(self):
        mock_market_book = mock.Mock()
        mock_runner = mock.Mock(selection_id=123, handicap=0)
        mock_market_book.runners = [mock_runner]
        self.assertEqual(utils.get_runner_book(mock_market_book, 123), mock_runner)

    @mock.patch("flumine.utils.get_price", return_value=1.01)
    def test_get_market_notes(self, mock_get_price):
        mock_market_book = mock.Mock()
        mock_runner = mock.Mock(selection_id=123, handicap=0, last_price_traded=5)
        mock_market_book.runners = [mock_runner]
        mock_market = mock.Mock(market_book=mock_market_book)
        self.assertEqual(utils.get_market_notes(mock_market, 123), "1.01,1.01,5")

    def test__get_event_ids(self):
        mock_markets = [
            mock.Mock(event_id=1, event_type_id="1", closed=False),
            mock.Mock(event_id=1, event_type_id="1", closed=False),
            mock.Mock(event_id=2, event_type_id="1", closed=False),
            mock.Mock(event_id=3, event_type_id="1", closed=True),
            mock.Mock(event_id=4, event_type_id="7", closed=False),
        ]
        self.assertEqual(utils.get_event_ids(mock_markets, "1"), [1, 2])



import logging

from .. import config
from ..markets.markets import Markets
from ..order.order import BaseOrder, OrderStatus, OrderTypes
from ..order.trade import Trade
from ..strategy.strategy import Strategies
from ..events.events import OrderEvent
from ..utils import STRATEGY_NAME_HASH_LENGTH

logger = logging.getLogger(__name__)

"""
Various functions to update current order status
and update status.

Loop through each current order:
    order = Lookup order in market using marketId and orderId
    if order is None (not present locally):
        create local order using data and make executable #todo!!!
    if betId is None (async placement):
        Update betId and log through logging_control
    if order betId != current_order betId:
        Get order using current_order betId due to replace request (new betId)
    if order:
        update current status
        process
"""


def process_current_orders(
    markets: Markets, strategies: Strategies, event, log_control
) -> None:
    for current_orders in event.event:
        for current_order in current_orders.orders:
            order_id = current_order.customer_order_ref[STRATEGY_NAME_HASH_LENGTH + 1 :]
            order = markets.get_order(
                market_id=current_order.market_id,
                order_id=order_id,
            )
            if order is None:
                logger.warning(
                    "Order %s not present in blotter" % current_order.bet_id,
                    extra={
                        "bet_id": current_order.bet_id,
                        "market_id": current_order.market_id,
                        "customer_strategy_ref": current_order.customer_strategy_ref,
                        "customer_order_ref": current_order.customer_order_ref,
                    },
                )
                order = create_order_from_current(markets, strategies, current_order)
                if order:
                    logger.info(
                        "Order %s added to blotter" % current_order.bet_id,
                        extra=order.info,
                    )
                    order.executable()  # todo correct?
                else:
                    continue

            if (
                config.async_place_orders
                and order.status == OrderStatus.PENDING
                and order.bet_id is None
                and current_order.bet_id
            ):  # async bet pending processing
                order.bet_id = current_order.bet_id
                log_control(OrderEvent(order))
                order.executable()
            elif order.bet_id != current_order.bet_id:  # replaceOrder handling (hacky)
                order = markets.get_order_from_bet_id(
                    market_id=current_order.market_id,
                    bet_id=current_order.bet_id,
                )

            if order:
                order.update_current_order(current_order)
                process_current_order(order)


def process_current_order(order: BaseOrder):
    if order.status == OrderStatus.EXECUTABLE:
        if order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            # todo use `order.current_order.status` ?
            if order.size_voided:
                order.voided()
            elif order.size_lapsed:
                order.lapsed()
            elif order.size_remaining == 0:
                order.execution_complete()
        elif order.order_type.ORDER_TYPE == OrderTypes.LIMIT_ON_CLOSE:
            if order.current_order.status == "EXECUTION_COMPLETE":
                order.execution_complete()
        elif order.order_type.ORDER_TYPE == OrderTypes.MARKET_ON_CLOSE:
            if order.current_order.status == "EXECUTION_COMPLETE":
                order.execution_complete()


def create_order_from_current(markets: Markets, strategies: Strategies, current_order):
    strategy_name_hash = current_order.customer_order_ref[:STRATEGY_NAME_HASH_LENGTH]
    order_id = current_order.customer_order_ref[STRATEGY_NAME_HASH_LENGTH + 1 :]
    # get market
    market = markets.markets.get(current_order.market_id)
    if market is None:
        logger.warning(
            "Market not available to create order {0}".format(order_id),
            extra={
                "bet_id": current_order.bet_id,
                "market_id": current_order.market_id,
                "customer_strategy_ref": current_order.customer_strategy_ref,
                "customer_order_ref": current_order.customer_order_ref,
                "strategy_name_hash": strategy_name_hash,
            },
        )
        return
    # get strategy
    strategy = strategies.hashes.get(strategy_name_hash)
    if strategy is None:
        logger.warning(
            "Strategy not available to create order {0}".format(order_id),
            extra={
                "bet_id": current_order.bet_id,
                "market_id": current_order.market_id,
                "customer_strategy_ref": current_order.customer_strategy_ref,
                "customer_order_ref": current_order.customer_order_ref,
                "strategy_name_hash": strategy_name_hash,
            },
        )
        return
    # add trade/order
    trade = Trade(
        market.market_id, current_order.selection_id, current_order.handicap, strategy
    )
    order = trade.create_order_from_current(current_order, order_id)
    market.blotter[order.id] = order
    return order



import logging
from tenacity import retry
from betfairlightweight import StreamListener
from betfairlightweight import BetfairError
from betfairlightweight.streaming.stream import BaseStream as BFBaseStream

from .basestream import BaseStream
from ..events.events import RawDataEvent
from ..exceptions import ListenerError

logger = logging.getLogger(__name__)

RETRY_WAIT = BaseStream.RETRY_WAIT


"""
Custom listener that doesn't do any processing,
helps reduce CPU.
"""


class FlumineListener(StreamListener):
    def _add_stream(self, unique_id: int, operation: str) -> BFBaseStream:
        if operation == "marketSubscription":
            return FlumineMarketStream(self, unique_id)
        elif operation == "orderSubscription":
            raise ListenerError("Unable to process order stream")
        elif operation == "raceSubscription":
            return FlumineRaceStream(self, unique_id)


class FlumineStream(BFBaseStream):
    def on_process(self, caches: list) -> None:
        output = RawDataEvent(caches)
        self.output_queue.put(output)

    def __str__(self):
        return "FlumineStream"

    def __repr__(self):
        return "<FlumineStream [%s]>" % len(self._caches)


class FlumineMarketStream(FlumineStream):

    _lookup = "mc"

    def _process(self, data: list, publish_time: int) -> bool:
        for market_book in data:
            market_id = market_book.get("id")
            if (
                "marketDefinition" in market_book
                and market_book["marketDefinition"]["status"] == "CLOSED"
            ):
                if market_id in self._caches:
                    # removes closed market from cache
                    del self._caches[market_id]
                    logger.info(
                        "[MarketStream: %s] %s removed, %s markets in cache"
                        % (self.unique_id, market_id, len(self._caches))
                    )
            elif self._caches.get(market_id) is None:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()
                logger.info(
                    "[MarketStream: %s] %s added, %s markets in cache"
                    % (self.unique_id, market_id, len(self._caches))
                )
            self._updates_processed += 1

        self.on_process([self.unique_id, publish_time, data])
        return False


class FlumineRaceStream(FlumineStream):

    _lookup = "rc"

    def _process(self, data: list, publish_time: int) -> bool:
        for update in data:
            market_id = update["mid"]
            if self._caches.get(market_id) is None:
                # adds empty object to cache to track live market count
                self._caches[market_id] = object()
                logger.info(
                    "[RaceStream: %s] %s added, %s markets in cache"
                    % (self.unique_id, market_id, len(self._caches))
                )
            self._updates_processed += 1

        self.on_process([self.unique_id, publish_time, data])
        return False


class DataStream(BaseStream):

    LISTENER = FlumineListener

    def __init__(self, *args, **kwargs):
        BaseStream.__init__(self, *args, **kwargs)
        self._listener = self.LISTENER(output_queue=self.flumine.handler_queue)

    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        logger.info(
            "Starting DataStream {0}".format(self.stream_id),
            extra={
                "stream_id": self.stream_id,
                "market_filter": self.market_filter,
                "market_data_filter": self.market_data_filter,
                "conflate_ms": self.conflate_ms,
            },
        )

        self._stream = self.betting_client.streaming.create_stream(
            unique_id=self.stream_id, listener=self._listener
        )
        try:
            self.stream_id = self._stream.subscribe_to_markets(
                market_filter=self.market_filter,
                market_data_filter=self.market_data_filter,
                conflate_ms=self.conflate_ms,
                initial_clk=self._listener.initial_clk,  # supplying these two values allows a reconnect
                clk=self._listener.clk,
            )
            self._stream.start()
        except BetfairError:
            logger.error(
                "DataStream {0} run error".format(self.stream_id), exc_info=True
            )
            raise
        except Exception:
            logger.critical(
                "DataStream {0} run error".format(self.stream_id), exc_info=True
            )
            raise
        logger.info("Stopped DataStream {0}".format(self.stream_id))



import queue
import logging
from betfairlightweight import BetfairError, filters
from tenacity import retry

from .basestream import BaseStream
from ..events.events import CurrentOrdersEvent
from .. import config

logger = logging.getLogger(__name__)

RETRY_WAIT = BaseStream.RETRY_WAIT


class OrderStream(BaseStream):
    @retry(wait=RETRY_WAIT)
    def run(self) -> None:
        logger.info(
            "Starting OrderStream {0}".format(self.stream_id),
            extra={
                "stream_id": self.stream_id,
                "customer_strategy_refs": config.hostname,
                "conflate_ms": self.conflate_ms,
                "streaming_timeout": self.streaming_timeout,
            },
        )
        if not self._output_thread.is_alive():
            logger.info(
                "Starting output_thread (OrderStream {0})".format(self.stream_id)
            )
            self._output_thread.start()

        self._stream = self.betting_client.streaming.create_stream(
            unique_id=self.stream_id, listener=self._listener
        )
        try:
            self.stream_id = self._stream.subscribe_to_orders(
                order_filter=filters.streaming_order_filter(
                    customer_strategy_refs=[config.hostname],
                    partition_matched_by_strategy_ref=True,
                    include_overall_position=False,
                ),
                conflate_ms=self.conflate_ms,
            )
            self._stream.start()
        except BetfairError:
            logger.error(
                "OrderStream {0} run error".format(self.stream_id), exc_info=True
            )
            raise
        except Exception:
            logger.critical(
                "OrderStream {0} run error".format(self.stream_id), exc_info=True
            )
            raise
        logger.info("Stopped OrderStream {0}".format(self.stream_id))

    def handle_output(self) -> None:
        """Handles output from stream."""
        while self.is_alive():
            try:
                order_books = self._output_queue.get(
                    block=True, timeout=self.streaming_timeout
                )
            except queue.Empty:  # todo snap every 5s or so anyway
                if self.flumine.markets.live_orders:
                    order_books = self._listener.snap(
                        market_ids=self.flumine.markets.open_market_ids
                    )
                else:
                    continue
            if order_books:
                self.flumine.handler_queue.put(CurrentOrdersEvent(order_books))

        logger.info("Stopped output_thread (OrderStream {0})".format(self.stream_id))



import datetime
import logging
import threading
from typing import Optional

from ..order.orderpackage import BaseOrder, OrderPackageType
from . import BaseControl
from ..clients.baseclient import BaseClient

logger = logging.getLogger(__name__)


class MaxTransactionCount(BaseControl):

    """
    Counts and limits orders based on max
    transaction count as per:
        - https://www.betfair.com/aboutUs/Betfair.Charges/#TranCharges2
        - 5000 transactions per hour
        - `A ‘transaction’ shall include all bets placed and all failed transactions`
    Counts are updated after an execution,
    thread safe due to the execution pool.
    """

    NAME = "MAX_TRANSACTION_COUNT"

    def __init__(self, flumine, client: BaseClient):
        super(MaxTransactionCount, self).__init__(flumine)
        self.client = client
        self._next_hour = None
        # this hour
        self.current_transaction_count = 0
        self.current_failed_transaction_count = 0
        # total since start
        self.transaction_count = 0
        self.failed_transaction_count = 0
        # thread lock
        self._lock = threading.Lock()

    def add_transaction(self, count: int, failed: bool = False) -> None:
        with self._lock:
            if failed:
                self.failed_transaction_count += count
                self.current_failed_transaction_count += count
            else:
                self.transaction_count += count
                self.current_transaction_count += count

    def _validate(self, order: BaseOrder, package_type: OrderPackageType) -> None:
        self._check_hour()
        if not self.safe:
            self._on_error(
                order,
                "Max Transaction Count has been reached ({0}) for current hour".format(
                    self.current_transaction_count_total
                ),
            )

    def _check_hour(self) -> None:
        if self._next_hour is None:
            self._set_next_hour()
        elif datetime.datetime.utcnow() > self._next_hour:
            logger.info(
                "Execution new hour",
                extra={
                    "current_transaction_count_total": self.current_transaction_count_total,
                    "current_transaction_count": self.current_transaction_count,
                    "current_failed_transaction_count": self.current_failed_transaction_count,
                    "total_transaction_count": self.transaction_count,
                    "total_failed_transaction_count": self.failed_transaction_count,
                    "client": self.client.info,
                },
            )
            self._set_next_hour()

    def _set_next_hour(self) -> None:
        now = datetime.datetime.utcnow()
        self._next_hour = (now + datetime.timedelta(hours=1)).replace(
            minute=0, second=0, microsecond=0
        )
        self.current_transaction_count = 0
        self.current_failed_transaction_count = 0

    @property
    def safe(self) -> bool:
        if self.transaction_limit is None:
            return True
        elif self.current_transaction_count_total <= self.transaction_limit:
            return True
        else:
            logger.error(
                "Transaction limit reached",
                extra={
                    "current_transaction_count_total": self.current_transaction_count_total,
                    "current_transaction_count": self.current_transaction_count,
                    "current_failed_transaction_count": self.current_failed_transaction_count,
                    "transaction_limit": self.transaction_limit,
                    "client": self.client.info,
                },
            )
            return False

    @property
    def current_transaction_count_total(self) -> int:
        return self.current_transaction_count + self.current_failed_transaction_count

    @property
    def transaction_count_total(self) -> int:
        return self.transaction_count + self.failed_transaction_count

    @property
    def transaction_limit(self) -> Optional[int]:
        return self.client.transaction_limit



import unittest
from unittest import mock

from flumine import FlumineBacktest
from flumine.clients import ExchangeType
from flumine.order.orderpackage import OrderPackageType
from flumine import config
from flumine.exceptions import RunError
from flumine.order.trade import TradeStatus
from flumine.markets.blotter import Blotter


class FlumineBacktestTest(unittest.TestCase):
    def setUp(self):
        self.mock_client = mock.Mock(EXCHANGE=ExchangeType.SIMULATED)
        self.flumine = FlumineBacktest(self.mock_client)

    def test_init(self):
        self.assertTrue(self.flumine.BACKTEST)

    def test_run_error(self):
        mock_client = mock.Mock()
        mock_client.EXCHANGE = 69
        self.flumine.client = mock_client
        with self.assertRaises(RunError):
            self.flumine.run()

    @mock.patch("flumine.backtest.backtest.FlumineBacktest._unpatch_datetime")
    @mock.patch("flumine.backtest.backtest.FlumineBacktest._process_end_flumine")
    @mock.patch("flumine.backtest.backtest.events")
    @mock.patch("flumine.backtest.backtest.FlumineBacktest._process_market_books")
    @mock.patch("flumine.backtest.backtest.FlumineBacktest._monkey_patch_datetime")
    def test_run(
        self,
        mock__monkey_patch_datetime,
        mock__process_market_books,
        mock_events,
        mock__process_end_flumine,
        mock__unpatch_datetime,
    ):
        mock_stream = mock.Mock(event_processing=False)
        mock_market_book = mock.Mock()
        mock_gen = mock.Mock(return_value=[[mock_market_book]])
        mock_stream.create_generator.return_value = mock_gen
        self.flumine.streams._streams = [mock_stream]
        self.flumine.run()
        mock__monkey_patch_datetime.assert_called_with()
        mock__process_market_books.assert_called_with(mock_events.MarketBookEvent())
        mock__process_end_flumine.assert_called_with()
        mock__unpatch_datetime.assert_called_with()

    @mock.patch("flumine.backtest.backtest.FlumineBacktest._unpatch_datetime")
    @mock.patch("flumine.backtest.backtest.FlumineBacktest._process_end_flumine")
    @mock.patch("flumine.backtest.backtest.events")
    @mock.patch("flumine.backtest.backtest.FlumineBacktest._process_market_books")
    @mock.patch("flumine.backtest.backtest.FlumineBacktest._monkey_patch_datetime")
    def test_run_event(
        self,
        mock__monkey_patch_datetime,
        mock__process_market_books,
        mock_events,
        mock__process_end_flumine,
        mock__unpatch_datetime,
    ):
        mock_stream_one = mock.Mock(event_processing=True, event_id=123)
        mock_market_book_one = mock.Mock(publish_time_epoch=321)
        mock_gen_one = mock.Mock(return_value=iter([[mock_market_book_one]]))
        mock_stream_one.create_generator.return_value = mock_gen_one

        mock_stream_two = mock.Mock(event_processing=True, event_id=123)
        mock_market_book_two = mock.Mock(publish_time_epoch=123)
        mock_gen_two = mock.Mock(return_value=iter([[mock_market_book_two]]))
        mock_stream_two.create_generator.return_value = mock_gen_two

        self.flumine.streams._streams = [mock_stream_one, mock_stream_two]
        self.flumine.run()
        mock__monkey_patch_datetime.assert_called_with()
        mock__process_market_books.assert_called_with(mock_events.MarketBookEvent())
        mock__process_end_flumine.assert_called_with()
        mock__unpatch_datetime.assert_called_with()

    @mock.patch("flumine.backtest.backtest.FlumineBacktest._process_backtest_orders")
    @mock.patch("flumine.backtest.backtest.FlumineBacktest._check_pending_packages")
    def test__process_market_books(
        self,
        mock__check_pending_packages,
        mock__process_backtest_orders,
    ):
        self.flumine.handler_queue.append(mock.Mock())
        mock_event = mock.Mock()
        mock_market_book = mock.Mock(market_id="1.23")
        mock_market_book.runners = []
        mock_market = mock.Mock(market_book=mock_market_book, context={})
        mock_market.blotter.live_orders = []
        self.flumine.markets._markets = {"1.23": mock_market}
        mock_event.event = [mock_market_book]
        self.flumine._process_market_books(mock_event)
        mock__check_pending_packages.assert_called_with("1.23")
        mock__process_backtest_orders.assert_called_with(mock_market)

    def test_process_order_package(self):
        mock_order_package = mock.Mock()
        self.flumine.process_order_package(mock_order_package)
        self.assertEqual(self.flumine.handler_queue, [mock_order_package])

    @mock.patch("flumine.backtest.backtest.process.process_current_order")
    def test__process_backtest_orders(self, mock_process_current_order):
        mock_market = mock.Mock(context={})
        mock_market.blotter = Blotter("1.23")
        mock_order = mock.Mock()
        mock_order.trade.status = TradeStatus.COMPLETE
        mock_order_two = mock.Mock()
        mock_order_two.trade.status = TradeStatus.COMPLETE
        mock_market.blotter._live_orders = [mock_order, mock_order_two]
        self.flumine._process_backtest_orders(mock_market)
        mock_process_current_order.assert_has_calls(
            [
                mock.call(mock_order),
                mock.call(mock_order_two),
            ]
        )
        self.assertEqual(mock_market.blotter._live_orders, [])

    def test__process_backtest_orders_strategies(self):
        mock_market = mock.Mock(context={})
        mock_market.blotter.live_orders = []
        mock_strategy = mock.Mock()
        self.flumine.strategies = [mock_strategy]
        self.flumine._process_backtest_orders(mock_market)
        mock_strategy.process_orders.assert_called_with(
            mock_market, mock_market.blotter.strategy_orders(mock_strategy)
        )

    def test__check_pending_packages_place(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23",
            package_type=OrderPackageType.PLACE,
            elapsed_seconds=5,
            bet_delay=1,
            client=mock_client,
            simulated_delay=1.2,
        )
        self.flumine.handler_queue = [mock_order_package]
        self.flumine._check_pending_packages("1.23")
        mock_client.execution.handler.assert_called_with(mock_order_package)

    def test__check_pending_packages_place_pending(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23",
            package_type=OrderPackageType.PLACE,
            elapsed_seconds=0.2,
            bet_delay=1,
            client=mock_client,
            simulated_delay=1.2,
        )
        self.flumine.handler_queue = [mock_order_package]
        self.flumine._check_pending_packages("1.23")
        mock_client.execution.handler.assert_not_called()

    def test__check_pending_packages_place_diff_market_id(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23",
            package_type=OrderPackageType.PLACE,
            elapsed_seconds=2,
            bet_delay=1,
            client=mock_client,
            simulated_delay=1.2,
        )
        self.flumine.handler_queue = [mock_order_package]
        self.flumine._check_pending_packages("1.24")
        mock_client.execution.handler.assert_not_called()

    def test__check_pending_packages_cancel(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23", elapsed_seconds=3, client=mock_client, simulated_delay=0.2
        )
        self.flumine.handler_queue = [mock_order_package]
        self.flumine._check_pending_packages("1.23")
        mock_client.execution.handler.assert_called_with(mock_order_package)

    def test__check_pending_packages_cancel_pending(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23", elapsed_seconds=2, client=mock_client, simulated_delay=0.2
        )
        self.flumine.handler_queue = [mock_order_package]
        self.flumine._check_pending_packages("1.23")
        mock_client.execution.handler.assert_called_with(mock_order_package)

    def test__check_pending_packages_update(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23", elapsed_seconds=3, client=mock_client, simulated_delay=0.2
        )
        self.flumine.handler_queue = [mock_order_package]
        self.flumine._check_pending_packages("1.23")
        mock_client.execution.handler.assert_called_with(mock_order_package)

    def test__check_pending_packages_update_pending(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23", elapsed_seconds=2, client=mock_client, simulated_delay=0.2
        )
        self.flumine.handler_queue = [mock_order_package]
        self.flumine._check_pending_packages("1.23")
        mock_client.execution.handler.assert_called_with(mock_order_package)

    def test__check_pending_packages_replace(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23",
            package_type=OrderPackageType.REPLACE,
            elapsed_seconds=5,
            bet_delay=1,
            client=mock_client,
            simulated_delay=1.2,
        )
        self.flumine.handler_queue = [mock_order_package]
        self.flumine._check_pending_packages("1.23")
        mock_client.execution.handler.assert_called_with(mock_order_package)

    def test__check_pending_packages_replace_pending(self):
        mock_client = mock.Mock()
        mock_order_package = mock.Mock(
            market_id="1.23",
            package_type=OrderPackageType.REPLACE,
            elapsed_seconds=2,
            bet_delay=1,
            client=mock_client,
            simulated_delay=1.2,
        )
        self.flumine.handler_queue.append(mock_order_package)
        self.flumine._check_pending_packages("1.23")
        mock_client.execution.handler.assert_called_with(mock_order_package)

    @mock.patch("flumine.baseflumine.BaseFlumine.info")
    @mock.patch("flumine.baseflumine.BaseFlumine.log_control")
    def test__process_close_market_closed(self, mock_log_control, mock_info):
        mock_strategy = mock.Mock()
        mock_strategy.stream_ids = [1, 2, 3]
        self.flumine.strategies = [mock_strategy]
        mock_market = mock.Mock(closed=False, elapsed_seconds_closed=None)
        mock_market.market_book.streaming_unique_id = 2
        mock_market.blotter.process_cleared_orders.return_value = []
        self.flumine.markets._markets = {
            "1.23": mock_market,
            "4.56": mock.Mock(market_id="4.56", closed=True, elapsed_seconds_closed=25),
            "7.89": mock.Mock(
                market_id="7.89", closed=True, elapsed_seconds_closed=3601
            ),
            "1.01": mock.Mock(
                market_id="1.01", closed=False, elapsed_seconds_closed=3601
            ),
        }
        mock_event = mock.Mock()
        mock_market_book = mock.Mock(market_id="1.23")
        mock_event.event = mock_market_book
        self.flumine._process_close_market(mock_event)
        self.assertEqual(len(self.flumine.markets._markets), 4)

    def test_str(self):
        assert str(self.flumine) == "<FlumineBacktest>"

    def test_repr(self):
        assert repr(self.flumine) == "<FlumineBacktest>"

    def test_enter_exit(self):
        control = mock.Mock()
        self.flumine._logging_controls = [control]
        self.flumine.simulated_execution = mock.Mock()
        self.flumine.betfair_execution = mock.Mock()
        with self.flumine:
            self.assertTrue(self.flumine._running)
            self.assertTrue(config.simulated)

        self.assertFalse(self.flumine._running)
        self.assertTrue(config.simulated)
        self.flumine.simulated_execution.shutdown.assert_called_with()
        self.flumine.betfair_execution.shutdown.assert_called_with()
        control.start.assert_called_with()

    def tearDown(self) -> None:
        config.simulated = False



import logging

from ..clients.clients import ExchangeType
from ..order.ordertype import OrderTypes
from ..order.orderpackage import OrderPackageType, BaseOrder
from . import BaseControl
from .. import utils

logger = logging.getLogger(__name__)


class OrderValidation(BaseControl):

    """
    Validates order price and size is valid for
    exchange.
    """

    NAME = "ORDER_VALIDATION"

    def _validate(self, order: BaseOrder, package_type: OrderPackageType) -> None:
        if order.EXCHANGE == ExchangeType.BETFAIR:
            self._validate_betfair_order(order)

    def _validate_betfair_order(self, order):
        if order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
            self._validate_betfair_size(order)
            self._validate_betfair_price(order)
            self._validate_betfair_min_size(order, OrderTypes.LIMIT)
        elif order.order_type.ORDER_TYPE == OrderTypes.LIMIT_ON_CLOSE:
            self._validate_betfair_price(order)
            self._validate_betfair_liability(order)
            self._validate_betfair_min_size(order, OrderTypes.LIMIT_ON_CLOSE)
        elif order.order_type.ORDER_TYPE == OrderTypes.MARKET_ON_CLOSE:
            self._validate_betfair_liability(order)
            self._validate_betfair_min_size(order, OrderTypes.MARKET_ON_CLOSE)
        else:
            self._on_error(order, "Unknown orderType")

    def _validate_betfair_size(self, order):
        if order.order_type.size is None:
            self._on_error(order, "Order size is None")
        elif order.order_type.size <= 0:
            self._on_error(order, "Order size is less than 0")
        elif order.order_type.size != round(order.order_type.size, 2):
            self._on_error(order, "Order size has more than 2dp")

    def _validate_betfair_price(self, order):
        if order.order_type.price is None:
            self._on_error(order, "Order price is None")
        elif utils.as_dec(order.order_type.price) not in utils.PRICES:
            self._on_error(order, "Order price is not valid")

    def _validate_betfair_liability(self, order):
        if order.order_type.liability is None:
            self._on_error(order, "Order liability is None")
        elif order.order_type.liability <= 0:
            self._on_error(order, "Order liability is less than 0")
        elif order.order_type.liability != round(order.order_type.liability, 2):
            self._on_error(order, "Order liability has more than 2dp")

    def _validate_betfair_min_size(self, order, order_type):
        client = self.flumine.client
        if client.min_bet_validation is False:
            return  # some accounts do not have min bet restrictions
        if order_type == OrderTypes.LIMIT:
            if (
                order.order_type.size < client.min_bet_size
                and (order.order_type.price * order.order_type.size)
                < client.min_bet_payout
            ):
                self._on_error(
                    order,
                    "Order size is less than min bet size ({0}) or payout ({1}) for currency".format(
                        client.min_bet_size, client.min_bet_payout
                    ),
                )
        else:  # todo is this correct?
            if (
                order.side == "BACK"
                and order.order_type.liability < client.min_bet_size
            ):
                self._on_error(
                    order,
                    "Liability is less than min bet size ({0}) for currency".format(
                        client.min_bet_size
                    ),
                )
            elif (
                order.side == "LAY"
                and order.order_type.liability < client.min_bsp_liability
            ):
                self._on_error(
                    order,
                    "Liability is less than min BSP payout ({0}) for currency".format(
                        client.min_bsp_liability
                    ),
                )


class StrategyExposure(BaseControl):

    """
    Validates:
        - `strategy.validate_order` function
        - `strategy.max_order_exposure` is not violated if order is executed
        - `strategy.max_selection_exposure` is not violated if order is executed

    Exposure calculation includes pending,
    executable and execution complete orders.
    """

    NAME = "STRATEGY_EXPOSURE"

    def _validate(self, order: BaseOrder, package_type: OrderPackageType) -> None:
        if package_type == OrderPackageType.PLACE:
            # strategy.validate_order
            runner_context = order.trade.strategy.get_runner_context(*order.lookup)
            if order.trade.strategy.validate_order(runner_context, order) is False:
                return self._on_error(order, order.violation_msg)

        if package_type in (
            OrderPackageType.PLACE,
            OrderPackageType.REPLACE,
        ):
            strategy = order.trade.strategy
            if order.order_type.ORDER_TYPE == OrderTypes.LIMIT:
                if order.side == "BACK":
                    order_exposure = order.order_type.size
                else:
                    order_exposure = (
                        order.order_type.price - 1
                    ) * order.order_type.size
            elif order.order_type.ORDER_TYPE == OrderTypes.LIMIT_ON_CLOSE:
                order_exposure = order.order_type.liability  # todo correct?
            elif order.order_type.ORDER_TYPE == OrderTypes.MARKET_ON_CLOSE:
                order_exposure = order.order_type.liability
            else:
                return self._on_error(order, "Unknown order_type")

            # per order
            if order_exposure > strategy.max_order_exposure:
                return self._on_error(
                    order,
                    "Order exposure ({0}) is greater than strategy.max_order_exposure ({1})".format(
                        order_exposure, strategy.max_order_exposure
                    ),
                )

            # per selection
            market = self.flumine.markets.markets[order.market_id]
            if package_type == OrderPackageType.REPLACE:
                exclusion = order
            else:
                exclusion = None

            current_exposures = market.blotter.get_exposures(
                strategy, lookup=order.lookup, exclusion=exclusion
            )
            """
            We use -min(...) in the below, as "worst_possible_profit_on_X" will be negative if the position is
            at risk of loss, while exposure values are always atleast zero.
            Exposure refers to the largest potential loss.
            """
            if order.side == "BACK":
                current_selection_exposure = -current_exposures[
                    "worst_possible_profit_on_lose"
                ]
            else:
                current_selection_exposure = -current_exposures[
                    "worst_possible_profit_on_win"
                ]
            potential_exposure = current_selection_exposure + order_exposure
            if potential_exposure > strategy.max_selection_exposure:
                return self._on_error(
                    order,
                    "Potential selection exposure ({0:.2f}) is greater than strategy.max_selection_exposure ({1})".format(
                        potential_exposure,
                        strategy.max_selection_exposure,
                    ),
                )



import logging
import queue
from threading import Thread

from ..events import events
from ..events.events import EventType

logger = logging.getLogger(__name__)


class LoggingControl(Thread):
    """
    Base Logging Control instance, used to output logging of
    trades, orders, new markets and market closure. Can be
    used to send data to a database or API for further analysis.
    """

    NAME = "LOGGING_CONTROL"

    def __init__(self, daemon: bool = True):
        Thread.__init__(self, daemon=daemon, name=self.NAME)
        self.logging_queue = queue.Queue()
        self.cache = []

    def run(self) -> None:
        logger.info("Starting logging control %s" % self.NAME)
        while True:
            event = self.logging_queue.get()
            if event is None:
                logger.info("Shutting down logging control %s" % self.NAME)
                break
            else:
                try:
                    self.process_event(event)
                except Exception as e:
                    logger.critical(
                        "{0} exception raised in {0}".format(e, self.NAME),
                        exc_info=True,
                        extra={"event": event},
                    )

    def process_event(self, event: events.BaseEvent):
        if event.EVENT_TYPE == EventType.CONFIG:
            self._process_config(event)

        elif event.EVENT_TYPE == EventType.STRATEGY:
            self._process_strategy(event)

        elif event.EVENT_TYPE == EventType.MARKET:
            self._process_market(event)

        elif event.EVENT_TYPE == EventType.TRADE:
            self._process_trade(event)

        elif event.EVENT_TYPE == EventType.ORDER:
            self._process_order(event)

        elif event.EVENT_TYPE == EventType.BALANCE:
            self._process_balance(event)

        elif event.EVENT_TYPE == EventType.CLEARED_ORDERS:
            self._process_cleared_orders(event)

        elif event.EVENT_TYPE == EventType.CLEARED_ORDERS_META:
            self._process_cleared_orders_meta(event)

        elif event.EVENT_TYPE == EventType.CLEARED_MARKETS:
            self._process_cleared_markets(event)

        elif event.EVENT_TYPE == EventType.CLOSE_MARKET:
            self._process_closed_market(event)

        elif event.EVENT_TYPE == EventType.CUSTOM_EVENT:
            self._process_custom_event(event)

        elif event.EVENT_TYPE == EventType.TERMINATOR:
            self._process_end_flumine(event)
            self.logging_queue.put(None)

        else:
            logger.error("Unwanted item in logging control: {0}".format(event))

    def _process_config(self, event):
        """
        :param event.event: config.py file
        """
        logger.debug("process_config: %s" % event)

    def _process_strategy(self, event):
        """
        :param event.event: Strategy class
        """
        logger.debug("process_strategy: %s" % event)

    def _process_market(self, event):
        """
        :param event.event: Market class
        """
        logger.debug("process_market: %s" % event)

    def _process_trade(self, event):
        """
        :param event.event: Trade class
        """
        logger.debug("process_trade: %s" % event)

    def _process_order(self, event):
        """
        :param event.event: Order class
        """
        logger.debug("process_order: %s" % event)

    def _process_balance(self, event):
        """
        :param event.event: Client class
        """
        logger.debug("process_balance: %s" % event)

    def _process_cleared_orders(self, event):
        """
        :param event.event: betfairlightweight resources.ClearedOrders
        """
        logger.debug("process_cleared_orders: %s" % event)

    def _process_cleared_orders_meta(self, event):
        """
        :param event.event: list of orders
        """
        logger.debug("process_cleared_orders_meta: %s" % event)

    def _process_cleared_markets(self, event):
        """
        :param event.event: betfairlightweight resources.ClearedOrders
        """
        logger.debug("process_cleared_markets: %s" % event)

    def _process_closed_market(self, event):
        """
        :param event.event: Market class
        """
        logger.debug("process_closed_market: %s" % event)

    def _process_custom_event(self, event):
        """
        :param event.event: Custom event
        """
        logger.debug("process_custom_event: %s" % event)

    def _process_end_flumine(self, event):
        """
        :param event.event: Termination Event
        """
        logger.debug("_process_end_flumine: %s" % event)
